"""
DenseNet for cifar with pytorch

Reference:
[1] H. Gao, Z. Liu, L. Maaten and K. Weinberger. Densely connected convolutional networks. In CVPR, 2017
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
from collections import OrderedDict
from .layers import *
from .param_bank import *

import math

class _DenseLayer(nn.Sequential):
    def __init__(self, bank, num_input_features, growth_rate, bn_size, drop_rate):
        super(_DenseLayer, self).__init__()
        self.add_module('norm1', nn.BatchNorm2d(num_input_features)),
        self.add_module('relu1', nn.ReLU(inplace=True)),
        if bank: self.add_module('conv1', SConv2d(bank, num_input_features, bn_size *
                        growth_rate, kernel_size=1, stride=1, bias=False)),
        else: self.add_module('conv1', nn.Conv2d(num_input_features, bn_size *
                        growth_rate, kernel_size=1, stride=1, bias=False)),
        self.add_module('norm2', nn.BatchNorm2d(bn_size * growth_rate)),
        self.add_module('relu2', nn.ReLU(inplace=True)),
        if bank: self.add_module('conv2', SConv2d(bank, bn_size * growth_rate, growth_rate,
                        kernel_size=3, stride=1, padding=1, bias=False)),
        else: self.add_module('conv2', nn.Conv2d(bn_size * growth_rate, growth_rate,
                        kernel_size=3, stride=1, padding=1, bias=False)),
        self.drop_rate = drop_rate

    def forward(self, x):
        new_features = super(_DenseLayer, self).forward(x)
        if self.drop_rate > 0:
            new_features = F.dropout(new_features, p=self.drop_rate, training=self.training)

        return torch.cat([x, new_features], 1)


class _DenseBlock(nn.Sequential):
    def __init__(self, bank, num_layers, num_input_features, bn_size, growth_rate, drop_rate):
        super(_DenseBlock, self).__init__()
        for i in range(num_layers):
            layer = _DenseLayer(bank, num_input_features + i * growth_rate, growth_rate, bn_size, drop_rate)
            self.add_module('denselayer%d' % (i + 1), layer)


class _Transition(nn.Sequential):
    def __init__(self, bank, num_input_features, num_output_features):
        super(_Transition, self).__init__()
        self.add_module('norm', nn.BatchNorm2d(num_input_features))
        self.add_module('relu', nn.ReLU(inplace=True))
        if bank: self.add_module('conv', SConv2d(bank, num_input_features, num_output_features,
                                                 kernel_size=1, stride=1, bias=False))
        else: self.add_module('conv', nn.Conv2d(num_input_features, num_output_features,
                                                kernel_size=1, stride=1, bias=False))
        self.add_module('pool', nn.AvgPool2d(kernel_size=2, stride=2))


class DenseNet_Cifar(nn.Module):
    r"""Densenet-BC model class, based on
    `"Densely Connected Convolutional Networks" <https://arxiv.org/pdf/1608.06993.pdf>`_

    Args:
        growth_rate (int) - how many filters to add each layer (`k` in paper)
        block_config (list of 4 ints) - how many layers in each pooling block
        num_init_features (int) - the number of filters to learn in the first convolution layer
        bn_size (int) - multiplicative factor for number of bottle neck layers
          (i.e. bn_size * k features in the bottleneck layer)
        drop_rate (float) - dropout rate after each dense layer
        num_classes (int) - number of classification classes
    """
    def __init__(self, share_type, upsample_type, num_templates, max_params, growth_rate=12, block_config=(16, 16, 16),
                 num_init_features=24, bn_size=4, drop_rate=0, num_classes=10, groups=None):

        super(DenseNet_Cifar, self).__init__()

        self.num_templates = num_templates
        self.bank = None
        if share_type != 'none':
            self.bank = ParameterGroups(groups, share_type, upsample_type, max_params, num_templates)

        # First convolution
        if self.bank: self.features = self.features = nn.Sequential(OrderedDict([
            ('conv0', SConv2d(self.bank, 3, num_init_features, kernel_size=3, stride=1, padding=1, bias=False)),
        ]))

        else: self.features = nn.Sequential(OrderedDict([
            ('conv0', nn.Conv2d(3, num_init_features, kernel_size=3, stride=1, padding=1, bias=False)),
        ]))

        # Each denseblock
        num_features = num_init_features
        for i, num_layers in enumerate(block_config):
            block = _DenseBlock(self.bank, num_layers=num_layers, num_input_features=num_features,
                                bn_size=bn_size, growth_rate=growth_rate, drop_rate=drop_rate)
            self.features.add_module('denseblock%d' % (i + 1), block)
            num_features = num_features + num_layers * growth_rate
            if i != len(block_config) - 1:
                trans = _Transition(self.bank, num_input_features=num_features, num_output_features=num_features // 2)
                self.features.add_module('transition%d' % (i + 1), trans)
                num_features = num_features // 2

        # Final batch norm
        self.features.add_module('norm5', nn.BatchNorm2d(num_features))

        # Linear layer
        if self.bank: self.classifier = SLinear(self.bank, num_features, num_classes)
        else: self.classifier = nn.Linear(num_features, num_classes)
        
        # initialize conv and bn parameters
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

        if self.bank:
            self.bank.setup_bank()

    def forward(self, x):
        features = self.features(x)
        out = F.relu(features, inplace=True)
        out = F.avg_pool2d(out, kernel_size=8, stride=1).view(features.size(0), -1)
        out = self.classifier(out)
        return out


def densenet_cifar(share_type, upsample_type, depth, k, num_templates, max_params, num_classes, groups, **kwargs):
    N = (depth - 4) // 6
    model = DenseNet_Cifar(share_type, upsample_type, num_templates, max_params, growth_rate=k, block_config=[N, N, N], num_init_features=2*k, num_classes=num_classes, groups=groups, **kwargs)
    return model


if __name__ == '__main__':
    net = densenet_BC_cifar(190, 40, num_classes=100)
    input = torch.randn(1, 3, 32, 32)
    y = net(input)
    print(net)
    print(y.size())
